/*  used to be
 *  1-Item score approximation software
 *  long-since derived from MFSIO.cp
 *  modified 9/2004 for IRTScore v. 2s for 285
 */

#include "IO.h"
using namespace std;

void GetControlNums(IRT08ss* test, ostream& OT) {

	char DatFile[64];  // note: maximum length of filename returned from SFGetFile is 64 bytes
	int i;
	int nitem;
	int maxcat = 2;
	int maxscore;
	bool nofile;

//	if (!OT) OT = &cout;

// Get parameter file
	do {
		nofile = true;
		cout << "Enter Input File Name: ";
		cin.getline(DatFile, 64);	
		ifstream dat (DatFile);
    	if (dat) nofile = false;    	
  		else cout << "No input file found; try again" << endl;
  	}
	while (nofile);
	
	(*test).ColItemCtr (DatFile, nitem);	// ColItemCtr is in IRT08ss
	OT << "Test has " << nitem << " items." << endl;
	(*test).ChangeNitems(nitem);
	strcpy((*test).DatFile, DatFile);	// DatFile is the filename -> test.DatFile
	(*test).ReadCol (DatFile, nitem, maxcat);	//ReadCol is in IRT08ss

	
	maxscore = 1;	// this & loop compute maximum score
	for (i = 0; i < nitem; i++) maxscore = maxscore + ((*test).I[i].ncat - 1);
	(*test).maxscore = maxscore;
	(*test).ChangeMaxScore(maxscore);
	(*test).recursive();	// this does the recursive algorithm

	WriteIt (OT, test);	// write out the tables
} // End of GetControlNums
	

// Writeit uses an OT file or cout to write the item parameters and score tables
void WriteIt (ostream& OT, IRT08ss* test) {
	
	OT << "Using parameters from " << (*test).DatFile << ":" << endl;

// print the item parameters
    OT << "Item Ncat Parameters:" << endl;
	OT << setprecision(2) << setiosflags(ios::right) 
	<< setiosflags(ios::showpoint | ios::fixed);
	for (int i = 0; i < (*test).nitem; i++) {
		OT << setw(4) << (i + 1) << setw(4) << (*test).I[i].ncat;
   		OT << setw(6) << (*test).I[i].a;
   		for (int j = 0; j < ((*test).I[i].ncat - 1); j++) 
   			OT << setw(6) << (*test).I[i].bj[j];
    	OT << endl;
	} // end of loop over items
	OT << endl;
	   
// print the score translation table
	int maxscore = (*test).maxscore;
	OT << setprecision(2) << setiosflags(ios::right)
		<< setiosflags(ios::showpoint | ios::fixed);
	OT << endl << "Score  EAP   SD    Prop.";
	OT << endl;
	for (int i = 0; i < maxscore; i++) {
		OT << setw(4) << i  
			<< setprecision(2) << setw(6) << (*test).eap[i] << setw(6) << (*test).sd[i]
			<< setprecision(5) << setw(9) << (*test).expd[i] << endl;
	}
	OT << endl;
} // end of WriteIt

// ScoreA ---
// computes response-pattern EAPs and MAPs and summed-score EAPs
// for user-entered response strings

void ScoreA(IRT08ss* test, ostream& OT) {
	char s[255];				// response patterns go in s
	char r[2];					// individual responses pass thru r
	valarray<int> x;					// x is the integer [0,1] responses
	valarray<double> posterior;			// response pattern posterior
	valarray<double> SSposterior;		// pattern of summed scores posterior
	valarray<double> MAPreturn(2);
	double expd;				// probability for resonse pattern
	double eap;					// EAP for response pattern
	double sd;					// posterior SD for response pattern
	double SSexpd;				// probability of summed score
	double SumScoreEap;			// summed score pattern EAP
	double SumScoreSd;			// summed score pattern posterior standard deviation
  	int i, t, totalscore;
  	bool inputNotOK;

//	if (!OT) OT = &cout;
	x.resize((*test).nitem);		// x is the integer [0,1] responses
	
	posterior.resize(IRT08ss::nqpt);
	SSposterior.resize(IRT08ss::nqpt);
// this do-while repeats indefinitely, until <CR> is entered with no input
	do {
// initialize anew for each person
		t = 0; // t is 0 for routing, 1 for 2nd stage, 2 for 3rd stage...
		posterior = IRT08ss::popdist; // initialize response pattern posterior to normal
		SSposterior = IRT08ss::popdist;	// then SS posterior

		totalscore = 0;

// input and check response string
		do {		
		inputNotOK = false;
			cout << "Enter response string, " 
				<< (*test).nitem << " graded item responses [<return> quits]:" << endl;
	
			cin.getline(s, 255);	// responses are read as alphabetic strings, so that
			if (strlen(s)) {		// carriage return makes strlen(s) false here
				for (i = 0; i < (*test).nitem; i++) {
					strncpy(r, &s[i], 1);
					x[i] = atoi(r);
				}
			}
			else return;
		
			totalscore = 0;
			
			if (strlen(s) != (*test).nitem) inputNotOK = true;
			
			for (i = 0; i < (*test).nitem; i++) {
				if (x[i] < 0) inputNotOK = true;
				if (x[i] >= (*test).I[i].ncat) inputNotOK = true;
				totalscore += x[i]; // compute the summed score
			}
			if (inputNotOK) {
				cout << "Either wrong number of responses, or" << endl;
				cout << "One or more responses out of range [0, # categories -1]; try again" << endl;
				cout << endl;
			}
			else cout << "Summed score: " << totalscore << endl;	
		} while (inputNotOK);	

// compute the pattern MAP
		MAPreturn = (*test).MAP(x);
		cout << setprecision(2) << setiosflags(ios::right)
			 << setiosflags(ios::showpoint | ios::fixed);
		cout << "Response pattern MAP =" 
			 << setw(6) << MAPreturn[0] << ", sd = " << MAPreturn[1] << endl;
				
// compute the pattern EAP
		for (i = 0; i < (*test).nitem; i++)
			posterior = posterior*(*test).I[i].T[x[i]]; // update
		(*test).int_statsR(posterior, expd, eap, sd); // pattern posterior
		cout << setprecision(2) << setiosflags(ios::right)
			 << setiosflags(ios::showpoint | ios::fixed);
		cout << "Response pattern EAP =" 
			 << setw(6) << eap << ", sd = " << sd << endl;
					
// now compute the pattern-of-summed-scores EAP
		SSposterior = SSposterior*(*test).lk[totalscore];	// update the stored summed-score posterior
		(*test).int_statsR(SSposterior, SSexpd, SumScoreEap, SumScoreSd);
		cout << "EAP given SumScore =" << setw(6) << SumScoreEap 
			 << ", sd = " << SumScoreSd << endl << endl;
					
	} while (strlen(s)); // end of indefinite do-while 
}

// SimB ---
// computes response-pattern EAPs and MAPs and summed-score EAPs
// for each simulee at each level of theta -- computes simulation summary statistics

void SimB(IRT08ss *test, ostream& OT) { 
	int s1;					// random number seed 1
	int s2;					// seed 2
	int nreps;				// number of simulees at each value of theta
	char s[255];			// user input goes in s
	valarray<double> posterior;		// response pattern posterior
	valarray<double> SSposterior;	// pattern of summed scores posterior
	valarray<double> MAPreturn(2);
	valarray<int> x((*test).nitem);	// x is the integer [0,1] responses
	double expd;				// probability for resonse pattern
	int theta_points = 30;		// set up here to simulate nreps simulees
	double low_theta = -2.9;	// @ each of 30 values of theta equispaced
	double theta_interval = 0.2;	// between -2.9 and 2.9

	posterior.resize(IRT08ss::nqpt);
	SSposterior.resize(IRT08ss::nqpt);
	
	int nests = 3; // 3 estimators, see description below
    int iPEAP = 0;
    int iPMAP = 1;
    int iSSEAP = 2;

//  The long names of each scoring method
    string Lstrings[3];
    Lstrings[iPEAP] = "Response pattern EAP";
    Lstrings[iPMAP] = "Response pattern MAP";
	Lstrings[iSSEAP] = "Summed score EAP";
	
//  The short names of each scoring method
    string Sstrings[3];
    Sstrings[iPEAP] = "P-EAP";
    Sstrings[iPMAP] = "P-MAP";
	Sstrings[iSSEAP] = "SS";
	
	valarray<double> prof(nests);		// prof holds estimates of proficiency
	valarray<double> stdev(nests);		// their standard deviations
	double SSexpd;				// probability of summed score
	int i, totalscore;

	time_t time1, time2;
	
	valarray<double> Ncount;
	
	vector< valarray<double> > bias(nests);	// theta_points is the number of 
	vector< valarray<double> > MSE(nests);	// values of theta
	vector< valarray<double> > var(nests);
	for (int est=0; est<nests; est++) {
		bias[est].resize(theta_points);	// theta_points is the number of 
		MSE[est].resize(theta_points);	// values of theta
		var[est].resize(theta_points);
	}
	
	double deviation;
	valarray<double> theta;
	valarray<double> NormalWts;
	double WtTotal;
	double Ptotal;
	double truethetass;
	
    valarray<double> corrs(nests);
    valarray<double> sumsqs(nests);
    valarray<double> bars(nests);
	valarray<double> AvgMSE(nests);
	int itheta;
	int simulees;
	double truetheta;
	bool badseed1, badseed2;
	
//	if (!OT) OT = &cout;

	do {
		badseed1 = false;
		cout << endl << "Enter seed1 (0 < seed1 < 31328) [return: default = 962]: ";
		cin.getline(s, 255);	
		if (strlen(s)) { 		
			s1 = atoi(s);
		}
		else s1 = 962;
		if ((s1 > 31328)||(s1 < 0)) {
			badseed1 = true;
			cout << endl << "seed1 must be between 0 and 31328; try again:" << endl;
			}
		}
	while (badseed1);
	
	do {
		badseed2 = false;
		cout << "Enter seed2 (0 < seed2 < 30081) [return: default = 5036]: ";
		cin.getline(s, 255);	
		if (strlen(s)) {
			s2 = atoi(s);
		}
		else s2 = 5036;
		if ((s2 > 30081)||(s2 < 0)) {
			badseed2 = true;
			cout << endl << "seed2 must be between 0 and 30081; try again:" << endl << endl;
			}
		}
	while (badseed2);
	 
	OT << "seed1 = " << s1 << "     seed2 = " << s2 << endl << endl;
	
	cout << "Enter number of replications @ each theta value [return: default = 1000]: ";	
	cin.getline(s, 255);	
	if (strlen(s)) {		
		nreps = atoi(s);
	}
	else nreps = 1000;
	OT << "Number of replications @ each theta value = " << nreps << endl << endl;

// initializations:
	Ptotal = 0;
	for (i = 1; i < nests; i++) {
		bars[i] = 0;
		sumsqs[i] = 0;
		corrs[i] = 0;
	}
	truethetass = 0;
	Ncount.resize(theta_points);
	theta.resize(theta_points);
	NormalWts.resize(theta_points);
	WtTotal = 0.0;

// statistics totaled over theta levels are normal-weighted, range -2.9 - 2.9
	for (itheta = 0; itheta < theta_points; itheta++) 
		theta[itheta] =  low_theta + itheta*theta_interval;
	NormalWts = exp(-(((theta-0.0)*(theta-0.0))/(2.0*1.0*1.0)));
	NormalWts = NormalWts/NormalWts.sum(); 
	time1 = time(0);

// the simulation computations begin here
	cout << "Computing at:" << endl;
	for (itheta = 0; itheta < theta_points; itheta++) {   //begin loop over theta levels
		truetheta = theta[itheta];
		cout << setiosflags(ios::showpoint | ios::fixed);
		cout << setprecision(1) << setw(6);
		cout << "theta: " << truetheta << endl;
		for (simulees = 0; simulees < nreps; simulees++) {   //begin loop over simulees
		// initialize anew for each person  		    		
			posterior = IRT08ss::popdist; // initialize response pattern posterior to normal
			SSposterior = IRT08ss::popdist; //then SS posterior	
			totalscore = 0;
				
			RandResponse(test, truetheta,  (*test).nitem, x, s1, s2);	
			totalscore = 0;
			
			// compute the summed score
			for (i = 0; i < (*test).nitem; i++) totalscore += x[i]; 
					
			//	compute the pattern EAP			
			for (i = 0; i < (*test).nitem; i++) 
				posterior = posterior*(*test).I[i].T[x[i]]; // product of trace lines		
			(*test).int_statsR(posterior, expd, prof[iPEAP], stdev[iPEAP]); // posterior

// compute the pattern MAP
			MAPreturn = (*test).MAP(x);
			prof[iPMAP] = MAPreturn[0];
			stdev[iPMAP] = MAPreturn[1];

			// compute the summed-score EAP
			SSposterior = SSposterior*(*test).lk[totalscore];	
			(*test).int_statsR(SSposterior, SSexpd, prof[iSSEAP], stdev[iSSEAP]);
				
			Ncount[itheta] += 1.0;

			// loop over each estimator to compute simulation summary statistics:
			for (i = 0; i < nests; i++) {
				deviation = prof[i] - truetheta;
				bias[i][itheta] += deviation;
				MSE[i][itheta] += deviation*deviation;
				var[i][itheta] += stdev[i]*stdev[i];
			}
			WtTotal = WtTotal + NormalWts[itheta];
			for (i = 0; i < nests; i++) {
				bars[i] += prof[i]*NormalWts[itheta];
				sumsqs[i] += prof[i]*prof[i]*NormalWts[itheta];
				corrs[i] += prof[i]*truetheta*NormalWts[itheta]; 
			}
			truethetass = truethetass + truetheta*truetheta*NormalWts[itheta];
	
		} // end of loop over individual simulees
	} // end of loop over thetas

	time2 = time(0);
	cout << setprecision(1) << setw(8);
	cout << "Time required: " << difftime(time2, time1) << " seconds" << endl << endl;

	truethetass = truethetass/WtTotal;
	for (i = 0; i < nests; i++) {
			AvgMSE[i] = 0;
			for (itheta = 0; itheta < theta_points; itheta++) {
				bias[i][itheta] = bias[i][itheta]/Ncount[itheta];
				MSE[i][itheta] = MSE[i][itheta]/Ncount[itheta];
				var[i][itheta] = var[i][itheta]/Ncount[itheta];
				AvgMSE[i] = AvgMSE[i] + NormalWts[itheta]*MSE[i][itheta];
			}
		sumsqs[i] = (sumsqs[i] - bars[i]*bars[i]/WtTotal)/WtTotal;
		bars[i] = bars[i]/WtTotal;
		corrs[i] = corrs[i] / sqrt(truethetass*WtTotal*sumsqs[i]*WtTotal);
	}

	OT << "Table of bias and MSE at values of theta:" << endl;
	OT << setprecision(6) << resetiosflags(ios::right) 
		<< resetiosflags(ios::showpoint | ios::fixed);
	OT << "Theta";
	 
	for (i = 0; i < nests; i++) 
		OT << "\t" << Sstrings[i] << "bias" << "\t" << Sstrings[i] << "MSE";		 
	OT << endl;
	
	for (itheta = 0; itheta < theta_points; itheta++) {
		OT << theta[itheta];
		for (i = 0; i < nests; i++) 				
			OT << "\t" << bias[i][itheta] << "\t" << MSE[i][itheta];
		OT << endl;
	}
	OT << endl;
	OT << "N(0,1) weighted:" << endl;
	OT << endl;
	
	for (i = 0; i < nests; i++) {
		OT << Lstrings[i] << ":" << endl;
		OT << "Average: " << bars[i] << " Variance: " << sumsqs[i] << endl;
		OT << "Average MSE: " << AvgMSE[i] << " sqrt: " << sqrt(AvgMSE[i]) << endl; 
		OT << "Squared correlation with theta: " << corrs[i]*corrs[i] << endl; 	 
		OT << endl;		
	} 
} // end of SimB

// Samples random responses given levels of theta (graded model version)
void RandResponse(IRT08ss* test, double theta, int nitems, valarray<int>& x, int s1, int s2) {
	int i,k;
	double T;	
	float RanRect;
	for (i = 0; i < nitems; i++) {
		x[i] = 0;
		RanRect = unif1(s1,s2);
		for (k = 0; k < ( (*test).I[i].ncat - 1); k++) {
			T = 1.0 / (1.0 + exp(-(*test).I[i].a*(theta - (*test).I[i].bj[k]))); 
			if (RanRect < T) x[i] = k + 1;
		}
	} 
}

/* This implements Knuth's Algorithm P (polar method) */
float norm(int s1, int s2) {
	float u1,u2,s,unif();
	static int val_ready = 0;
	static float x1,x2;
   
	if(val_ready) {
		val_ready = 0;
		return x2; 
	}
	else {
		val_ready = 1;
		do {
			u1 = 2.0*unif1(s1,s2) - 1.0;
			u2 = 2.0*unif1(s1,s2) - 1.0;
			s = u1*u1 + u2*u2;
			if(s < 1.0) {
				x1 = u1*sqrt(-2.0*log(s)/s);
				x2 = u2*sqrt(-2.0*log(s)/s);
				return x1; 
			}
		} while(s >= 1.0); 
	} return -99.0;
}

/* Uniform RNG based on the Marsaglia algorithm */
float unif1(int s1, int s2) {
	float uni,s,t;
	static float u[97],c,cd,cm;
	int i,j,k,l,m,ii,jj;
	static int i97,j97;
	static int initialized = 0;

	/* Initialize the RNG */
	if ( !initialized ) {
		/* Check for legal seeds */
		if ((s1<0||s1>31328)||(s2<0||s2>30081)) {
			cout << "The first random number seed"
			" must have a value between 0 and 31328\n" << endl;
			cout << "The second random number"
			" seed must have a value between 0 and 30081\n" << endl;
			exit(1); 
		}

		i = (int)fmod(s1/177.0, 177.0) + 2;
		j = (int)fmod(s1      , 177.0) + 2;
		k = (int)fmod(s2/169.0, 178.0) + 1;
		l = (int)fmod(s2      , 169.0);
		for (ii=0;ii<97;ii++) {
			s = (float)0.0;
			t = (float)0.5;
			for (jj=0;jj<24;jj++) {
				m = (int)fmod( fmod(i*j,179.0)*k , 179.0 );
				i = j;
				j = k;
				k = m;
				l = (int)fmod( 53.0*l+1.0 , 169.0 );
				if ( fmod(l*m,64.0) >= 32)
					s += t;
				t = (float)(0.5 * t); 
			}
			u[ii] = s; 
		}
		c  = (float)(  362436.0 / 16777216.0);
		cd = (float)( 7654321.0 / 16777216.0);
		cm = (float)(16777213.0 / 16777216.0);
		i97 = 96;
		j97 = 32;

		initialized = 1; 
	}

	/* Get a random number. */
	uni = u[i97] - u[j97];
	if (uni < 0.0F) uni += 1.0;
	u[i97] = uni;
	i97--;
	if (i97 < 0) i97 = 96;
	j97--;
	if (j97 < 0) j97 = 96;
	c -= cd;
	if (c < 0.0F) c += cm;
	uni -= c;
	if (uni < 0.0F) uni += 1.0;
	return uni; 
}
