#include "IRT08ss.h"
using namespace std;

#define kZeroByte 0 // this is used in the code to set the null bit that ends C strings

extern char EAPname[];
int IRT08ss::size = 1;		// default size of IRT08sss is 1
int IRT08ss::nqpt = 49;		// default # quad pts is 46
double IRT08ss::qpmax = 6.0;	// also static; quadrature is set
								// to -6.0, -5.75, -5.5, ... 6.0 to match current IRTPRO defaults
valarray<double> IRT08ss::popdist;	// population distribution as a list, and
valarray<double> IRT08ss::theta;	// values of theta popdist and T, etc., are computed for

//default constructor
IRT08ss::IRT08ss() {
	int i;
	
	// what follows is a kluge to try to learn to set the statics
	// without making them global
	IRT08ss::popdist.resize(IRT08ss::nqpt);
	IRT08ss::theta.resize(IRT08ss::nqpt);

	// the following six lines create the lists of IRT08ss::theta (quadrature points)
	// and the ordinates of the normal distribution at those points
	double incr = (2 * qpmax)/(nqpt-1);
	IRT08ss::theta[0] = -qpmax;
	for (i = 1; i < nqpt; i++) 
		IRT08ss::theta[i] = IRT08ss::theta[i-1] + incr;
	IRT08ss::popdist = exp(-(((IRT08ss::theta-0.0)*(IRT08ss::theta-0.0))/(2.0*1.0*1.0)));
	IRT08ss::popdist = IRT08ss::popdist/IRT08ss::popdist.sum(); 

	nitem = size;
	maxscore = size + 1;
	for (i = 0; i < 64; i++)
		DatFile[i] = kZeroByte;
	I = new ItemStuff [nitem]; // later redo this as a vector?
	lk.resize(maxscore); 

	aCol = 1;
}

// destructor
IRT08ss::~IRT08ss() {
    delete [] I;
}

// sets the default constructor size, used when making arrays of these
static void setnitem(int n) {
  	IRT08ss::size = n;
}

// resize IRT08ss
void IRT08ss::ChangeNitems(int nn) {
    delete [] I;
    int i;
	nitem = nn;
	for (i = 0; i < 64; i++)
		DatFile[i] = kZeroByte;
	I = new ItemStuff [nitem];
}

// resize the output parts of IRT08ss that need to know maxscore to size
void IRT08ss::ChangeMaxScore(int m) {
//	delete [] lk;
	maxscore = m;
	eap.resize(maxscore);	// maxscore is actual length of these lists,
	sd.resize(maxscore);	// so its value is one greater than numerical
	expd.resize(maxscore);
	lk.resize(maxscore);
}  

// graded: input is the item index number i, telling where in IRT08ss this goes;
// the slope a and the DblList b containing the b parameters
// (the number of categories is assumed to be in test.I[i].ncat)
// as output, it places in test.I[i].T[x] a DblList of each
// trace line's ordinate at each of the values of IRT08ss::theta
void IRT08ss::graded (int i) {
	for  (int k=0; k<I[i].ncat; k++) I[i].T[k].resize(IRT08ss::nqpt);
	I[i].T[0] = 1.0 - (1.0 / (1.0 + exp(-I[i].a*(IRT08ss::theta - I[i].bj[0]))));
	for (int j = 1; j < I[i].ncat - 1; j++)
		 I[i].T[j] = (1.0 / (1.0 + exp(-I[i].a*(IRT08ss::theta - I[i].bj[j-1]))))
		 		- (1.0 / (1.0 + exp(-I[i].a*(IRT08ss::theta - I[i].bj[j]))));
		 I[i].T[I[i].ncat - 1] =
		 		1.0 / (1.0 + exp(-I[i].a*(IRT08ss::theta - I[i].bj[I[i].ncat - 2])));
} // end of graded

// ColItemCtr counts the number of items for a parameters-in-columns file in filename[]
void IRT08ss::ColItemCtr (char filename[], int& items) {
	char colline[256];	// this is set up for 255 columns plus endline
  	ifstream col2 (filename);
	items = 0;
// loop over items, read until EOF:
	while (col2.getline(colline, 256, '\n')) {
		items++;
	}
} //end of ColItemCtr

//  ReadCol for graded -- gets parameters from file in column format
void IRT08ss::ReadCol (char filename[], int items, int maxcat) {
	char colline[256];	// this is set up for 255 columns plus endline character
	char col[256];
	char* colstar[256];
	int ncols = 0;
	char s[255];
	int ncat;
	double a;	//Graded a's
	valarray<double> b;	//Graded b's
  	int i, k;
	ifstream col2 (filename);
	
	cout << "Enter number of categories: ";			
	cin.getline(s, 255);
	if (strlen(s)) {
		istringstream theAnswer(s);
		theAnswer >> ncat;
	}
	maxcat = ncat;
	bColList.resize(ncat - 1);
	cout << "Column for a?: ";
	cin.getline(s, 255);
	if (strlen(s)) {
		istringstream theResponse2(s);
		theResponse2 >> aCol;
	}
	for (i = 0; i < ncat - 1; i++) {	// loop to allow for graded b 
		cout << "Column for b" << i + 1 << "?: ";			
		cin.getline(s, 255);
		if (strlen(s)) {					
			istringstream theResponse3(s);		// theResponse3 and bCol need to be indexed for graded
			theResponse3 >> bColList[i];	
		} 
	}

// loop over items
	for (i=0; i < items; i++) {
		col2.getline(colline, 256, '\n');
		I[i].ChangeSize(ncat); 
 		b.resize(ncat-1);
		istringstream theLine(colline);
		while (theLine >> col) ncols++;
		istringstream theLine2(colline);
		for (int ii = 0; ii < ncols; ii++) {
			colstar[ii] = new char[256];
			theLine2 >> colstar[ii];
		} 	 
		a = atof(colstar[aCol-1]);
		for (k = 0; k < (ncat-1); k++)
			b[k] = atof(colstar[bColList[k]-1]);
		for (int ii = 0; ii < ncols; ii++) 
			delete [] colstar[ii];
		I[i].a = a;
		I[i].bj = b;
		graded(i);	// graded computes the actual trace lines 
  	} // end of loop over items
} // end of ReadCol

// int_stats: close copy of the LISP function of the same name---
// input is a valarray representing posterior likelihood,
// the IRT08ss::theta valarray,
// and a summed score that indicates where in the output the results go---
// computes EAP, S.D., and P for the posterior and places them in score-indexed locations...
// and that's all post 8/2010 revision
void IRT08ss::int_stats(valarray<double>& posterior, int score) {
	expd[score] = posterior.sum();
	eap[score] = (posterior * IRT08ss::theta).sum() / expd[score];
	sd[score] = sqrt((posterior * 
		(IRT08ss::theta - eap[score]) * (IRT08ss::theta - eap[score])).sum()/expd[score]);
} // end of int_stats

// int_statsR--this is a simpler version that just computes the statistics
// and returns them
void IRT08ss::int_statsR(valarray<double>& posterior, 
					   double& expdvalue, double& eapvalue, double& sdvalue) {
	expdvalue = posterior.sum();
	eapvalue = (posterior * IRT08ss::theta).sum() / expdvalue;
	sdvalue = sqrt((posterior * 
		(IRT08ss::theta - eapvalue) * (IRT08ss::theta - eapvalue)).sum()/expdvalue);
} // end of int_stats

// a close clone of the LISP we did (not)
// publish in the '95 APM article; it's in the Lab text report
void IRT08ss::recursive () {
	int i, ii, k, kk, j;
	vector< valarray<double> > zero_list;
	vector< valarray<double> > lk_temp;
	zero_list.resize(maxscore);
	lk_temp.resize(maxscore);
	lk.resize(maxscore);
	for (int s=0; s<maxscore; s++) {
		zero_list[s].resize(IRT08ss::nqpt);
		zero_list[s] = 0.0;
		lk_temp[s].resize(IRT08ss::nqpt);
		lk[s].resize(IRT08ss::nqpt);
	}
	for (i = 0; i < I[0].ncat; i++) lk[i] = I[0].T[i];
	int I_star = I[0].ncat;

	// Begin recursive algorithm
	for (ii = 1; ii < nitem; ii++) {
		for (kk = 0; kk < (I_star + I[ii].ncat - 1); kk++) 
			lk_temp[kk] = zero_list[kk];   
		for (k = 0; k < I[ii].ncat; k++) {
			for (j = 0; j < I_star; j++)
				lk_temp[j + k] = (lk[j] * I[ii].T[k]) + lk_temp[j + k];
		}
		I_star = I_star + (I[ii].ncat - 1);
		for (j = 0; j < I_star; j++) lk[j] = lk_temp[j]; 
	} // the recursive algorithm ends here
	
// below is stuff added to pick up eaps, weights, etc.
	valarray<double> posterior;
	posterior.resize(IRT08ss::nqpt);
	for (i = 0; i < maxscore; i++) {
		posterior = lk[i] * IRT08ss::popdist;
		int_stats(posterior, i);
	} // end of loop over scores
} // end of recursive


valarray<double> IRT08ss::MAP(valarray<int>& x) {
	valarray<double> values(2);
	double thetahat;
	double Tvalue;
	double l, dl, d2l;
	double StepLimit;
	double Lastdl;
	double correction;
	double p1,p2,q1,q2,p1q1,p2q2,dlcl;
	int i, iter, maxiter;
	
	StepLimit = 0.5;
	maxiter = 20;
	
	Lastdl = 0.0;
	thetahat = 0;

	for (iter = 0; iter < maxiter; iter++) {
		l = 0;
		dl = 0;
		d2l = 0;

		for (i = 0; i < nitem; i++) {
// a skip for missing goes here and is not here	
			if (x[i] == 0) {
				p1 = 1.0;
				p2 = 1.0 / (1 + exp(-I[i].a*(thetahat - I[i].bj[0])));
			} else {
				if (x[i] == (I[i].ncat - 1)) {
					p2 = 0.0;
				 	p1 = 1.0 / (1.0 + exp(-I[i].a*(thetahat - I[i].bj[I[i].ncat - 2])));
				} else {
					p1 = 1.0 / (1.0 + exp(-I[i].a*(thetahat - I[i].bj[x[i]-1])));
		 			p2 = 1.0 / (1.0 + exp(-I[i].a*(thetahat - I[i].bj[x[i]])));
				} 
			}
			Tvalue = p1 - p2;
			q1 = 1.0 - p1;
			q2 = 1.0 - p2;
			p1q1 = p1*q1;
			p2q2 = p2*q2;
     		dl = dl + (I[i].a*(p1q1 - p2q2) / Tvalue);
    		dlcl = (q1-p1)*p1q1+(p2-q2)*p2q2-((p2q2-p1q1)*(p2q2-p1q1)/Tvalue);
     		d2l = d2l + dlcl*I[i].a*I[i].a/Tvalue;
			l = l + log(Tvalue);
		}

// population distribution hard-coded as N(0,1) here
		l = l - 0.5*thetahat*thetahat;
		dl = dl - thetahat;
		d2l = d2l - 1;

		if (d2l > -0.01) d2l = -0.01;
		correction = dl/d2l;
		if (abs(correction) > StepLimit) 
			if (correction > 0) correction = StepLimit;
			else correction = -StepLimit;
		thetahat = thetahat - correction;
		if (abs(correction) < 0.000001) break;
		if ((dl*Lastdl) < 0.0) {
			thetahat = thetahat + (0.5*correction);
			StepLimit = 0.5*StepLimit;
			dl = 0.0;
		}
		Lastdl = dl;
	}
	values[0] = thetahat;
	values[1] = sqrt(-1.0/d2l);
	return values;
}


